/*
 * eps_restart_log_ota
 * 
 * Shows how to use ESPhttpUpdate to download firmware OTA from a
 * http server in case a watch dog expired or an exception occured.
 * This version sends message to a syslog server using the
 * Syslog library by Martin Sloup https://github.com/arcao/Syslog.git
 * 
 */
 
#include <ESP8266WiFi.h>
#include <ESP8266httpUpdate.h>
#include <mdEspRestart.h>
#include <Syslog.h>
#include <WiFiUdp.h>

/*
 * include one and only one of the following
 * language files
 */
#include "esp_boot_ota_en.h"
//#include "esp_boot_ota_fr.h"

extern "C" {
#include "user_interface.h"
}

#define VERSION     0x000102   // version 0.1.2


// -- Logging -- Combined Syslog and serial output --

#define VERBOSE_NONE  0  // nothing is sent to the serial port
#define VERBOSE_ECHO  1  // whatever is sent to syslog is sent to the serial port
#define VERBOSE_ALL   2  // everything is sent to the serial port

uint8_t verbose       = VERBOSE_ALL;   // set to VERBOSE_ALL and start serial monitor to see how this works
uint16_t logThreshold = LOG_DEBUG;     // usually set to LOG_ERR, but LOG_DEBUG will send all log messages to server
                                   // Simple way to view log messages on the log server 
                                   //     pi@raspberrypi $ tail -f /var/log/syslog    
/*
  Log priorities, defined in Syslog.h
  
  #define LOG_EMERG 0   // system is unusable 
  #define LOG_ALERT 1   // action must be taken immediately 
  #define LOG_CRIT  2   // critical conditions 
  #define LOG_ERR   3   // error conditions 
  #define LOG_WARNING 4 // warning conditions 
  #define LOG_NOTICE  5 // normal but significant condition 
  #define LOG_INFO  6   // informational 
  #define LOG_DEBUG 7   // debug-level messages 
*/
 
// Syslog server connection info
#define SYSLOG_SERVER "domo.local" //"my_log_server.local"   // adjust 
#define SYSLOG_PORT 514

// This device info
#define DEVICE_HOSTNAME "ESP8266"
#define APP_NAME "my_app:"
#define LOG_FACILITY LOG_USER    // The facility represents the machine process that created the syslog event
                                  
/* 
 Facility codes, defined in Syslog.h
   
  #define LOG_KERN  (0<<3)  // kernel messages
  #define LOG_USER  (1<<3)  // random user-level messages
  #define LOG_MAIL  (2<<3)  // mail system
  #define LOG_DAEMON  (3<<3)  // system daemons
  #define LOG_AUTH  (4<<3)  // security/authorization messages
  #define LOG_SYSLOG  (5<<3)  // messages generated internally by syslogd
  #define LOG_LPR   (6<<3)  // line printer subsystem
  #define LOG_NEWS  (7<<3)  // network news subsystem
  #define LOG_UUCP  (8<<3)  // UUCP subsystem
  #define LOG_CRON  (9<<3)  // clock daemon
  #define LOG_AUTHPRIV  (10<<3) // security/authorization messages (private)
  #define LOG_FTP   (11<<3) // ftp daemon
  // codes 12 through 15 reserved for system use
  #define LOG_LOCAL0  (16<<3) // reserved for local use
  #define LOG_LOCAL1  (17<<3) // reserved for local use
  #define LOG_LOCAL2  (18<<3) // reserved for local use
  #define LOG_LOCAL3  (19<<3) // reserved for local use
  #define LOG_LOCAL4  (20<<3) // reserved for local use
  #define LOG_LOCAL5  (21<<3) // reserved for local use
  #define LOG_LOCAL6  (22<<3) // reserved for local use
  #define LOG_LOCAL7  (23<<3) // reserved for local use
*/

// A UDP instance to let us send and receive packets over UDP
WiFiUDP udpClient;

// Create a syslog instance
Syslog syslog(udpClient, SYSLOG_SERVER, SYSLOG_PORT, DEVICE_HOSTNAME, APP_NAME, LOG_FACILITY);

/* 
 * Use setLogThreshold to set the messages that will be sent on to the syslog server. 
 * All messages with a priority level less than 'level' are to 
 * be logged also.
 * 
 */
void setLogThreshold(uint16_t threshold) {
  logThreshold = threshold;
  syslog.logMask(LOG_UPTO(threshold));
  Serial.printf("threshold: %d, logthreshold: %d", threshold, logThreshold);
}


/*
 * The function that sends the message (msg) to the syslog server if 'level' is
 * less than or equal to the priority level set with setlogThreshold. The
 * message will be echoed to the serial connection if verbose = VERBOSE_ALL.
 * If verbose = VERBOSE_ECHO, msg will be send to the serial connection only
 * if msg is actually sent to the syslog server.
 * 
 */
void log(uint16_t level, const  char *msg) {
  Serial.printf("log(%d, %s) with logThreshold: %d\n", level, msg, logThreshold);
  if (((verbose) && (level <= logThreshold)) || (verbose == 2)) {
    Serial.println(msg);
  }
  syslog.log(level, msg);
}

/*
 * Formated version of function log().
 */
__attribute__((format (printf, 2, 3))) 
void logf(uint16_t level, const char *format, ...)  {
  va_list args;
  char line[250];
  va_start(args, format);
  vsnprintf(line, 250, format, args);
  va_end(args);
  log(level, line);
} 


// -- Wi-Fi ------------

#define SSID_SZ     33  // maximum size of SSID
#define PWD_SZ      65  // maximum size of password minimum 8

/* 
 * Fallback WiFi network credentials
 * 
 * If a WiFi connection is not established when the ESP starts,
 * then these credentials will be used.
 * 
 */
#define NET_SSID    "" 
#define NET_PSK     "" 

/*
 * WiFi network credentials for the http server that can serve firmware OTA.
 * The server does not have to be on the same WiFi network as used by
 * the ESP.
 * 
 * When trying to do a firmware update, the ESP uses the following sequence
 * to set the WiFi credentatials:
 * 
 *   1) OTA_SSID if defined 
 *   2) WiFi.SSID() if defined (the current WiFi connection)
 *   3) NET_SSID if defined
 *   
 * If no SSID is define then OTA update will fail.   
 */
#define OTA_SSID    "" 
#define OTA_PSK     "" 

/*
 * URL of the firmware to upload OTA if necessary
 * Can be IP address or a domain name including
 * xxxxx.local zeroconf domains.
 */
#define AUTO_URL "http://192.168.1.22/ota/test.bin"
//#define AUTO_URL "http://haserver.local/ota/test.bin"


/*
 * Stops ESP when firmware has just been downloaded over UART 
 * telling the user to perform a hardware reset. Otherwise
 * the ESP will freeze (remain in programming mode) when
 * reset. This will not be needed with a Lolin/Wemos Mini D1 
 * and perhaps other boards that perform a hard reset after 
 * flashing the firmware.
 * 
 * It would be best to have a visual indication of this state, 
 * by quickly blinking LED in the while() loop for example.
 */
void stopUartMode(void) {
  if (isUartBoot()) {
    for (int i = 0; i < 3; i++) {
      log(LOG_ERR, stopUartModeString[i]);
    }
    while (1) {  yield(); }
  }
}


void showWiFiStatus(void) {
  if (WiFi.status() == WL_CONNECTED) {
    logf(LOG_DEBUG, S_WIFI_SSID_IP, WiFi.SSID().c_str(), WiFi.localIP().toString().c_str());
  } else {  
    log(LOG_ERR, S_WIFI_NOT_CONNECTED);
  }
}

/*
 * Define the time allowed for establishing a Wi-Fi connection
 * in 100's ms (1/10 seconds)
 */
#define CONNECT_TIME 1200  // 1200*100 = 120 000 ms = 2 minutes for WiFi connection

/*
 * Connects to a Wi-Fi network with the specified SSID and password (which can
 * be empty). Used to connect to working WiFi network and OTA WiFi network.
 * 
 * Can be called with an empty ssid, in which case, the last used SSID will be
 * used.
 * 
 * Returns false if
 * 
 *   - the ssid is empty and the last used SSID is empty
 *   - unable to put WiFi radio in station mode
 *   - the ssid does not correspond to an avaiable network
 *   - the password is not correct 
 *   - the connection was not established within the specified timeout
 *   
 * Returns true if  
 * 
 *   - the WiFi is already connected to the specified ssid or and ssid = "" and WiFi is already connected,
 *   - a new WiFi connection was established.
 *   
 */
boolean connectWiFi(char * ssid, char * pswd, int connectTime = CONNECT_TIME) {

  if ((strlen(ssid) == 0) && (WiFi.SSID().length() == 0)) {
    log(LOG_ERR, S_NET_SSID_NOT_DEFINED);
    return false;
  }

  if ( WiFi.isConnected() && ( !strlen(ssid)  ||  (!strcmp(ssid, WiFi.SSID().c_str())) ) ) {
    logf(LOG_INFO, S_ALREADY_CONNECTED, ssid);
    return true;
  }
  
  for (int i=0; i < 4; i++) {
    if (WiFi.mode(WIFI_STA)) break;
    if (i >= 3) {
      log(LOG_ERR, S_UNABLE_TO_SET_WIFI_STA);
      return false;
    }  
    delay(100);
  }

  if ( strlen(ssid) ) {
    logf(LOG_DEBUG, S_ATTEMPT_CONNECT, ssid);
    WiFi.begin(ssid, pswd);
  } else {
    logf(LOG_DEBUG, S_ATTEMPT_CONNECT, WiFi.SSID().c_str());
    WiFi.begin();    
  }

  /*
   * This is similar to WiFi.waitForConnectResult() except
   * for an (almost) assured exit result:  connected or not 
   * and if not connected because of bad ssid, bad
   * password or timedout.
   */
  for (int i = 0; i < connectTime; i++) {
    delay(100); // 
    switch (WiFi.status()) {
      case WL_CONNECTED: 
        logf(LOG_DEBUG, S_CONNECTED_WIFI, WiFi.SSID().c_str());
        return true;
        break;
        
      case WL_NO_SSID_AVAIL: 
        logf(LOG_ERR, S_NET_NOT_FOUND,  WiFi.SSID().c_str());
        return false;
        break;
        
      case WL_CONNECT_FAILED: 
        log(LOG_ERR, S_BAD_PASSWORD); 
        return false;
        break;
    }
  }
  log(LOG_ERR, S_NET_CONNECT_TIMEOUT);
  return false;
}


boolean performOtaUpdate(const char *url) {

  ESPhttpUpdate.rebootOnUpdate(false); // we will take care of reboot after some housekeeping
  switch(ESPhttpUpdate.update(url)) {
    case HTTP_UPDATE_FAILED:
      logf(LOG_ERR, S_HTTP_UPDATE_FAILED, ESPhttpUpdate.getLastError(), ESPhttpUpdate.getLastErrorString().c_str());
      return false;
      break;
  
    case HTTP_UPDATE_NO_UPDATES:
      log(LOG_ERR, S_HTTP_UPDATE_NO_UPDATES);
      return false;
      break;
  
    case HTTP_UPDATE_OK:
      log(LOG_DEBUG, S_HTTP_UPDATE_OK);
      return true;
      break;
  }
  log(LOG_DEBUG, S_OTA_UNKNOWN_STATE);
  return false;
}

/*
 * Usually WiFi is in station mode and autoconnect is true. 
 * Eventhough otaUpdate is called by setup() before the
 * WiFi.begin() instruction, it must be assumed that
 * WiFi could be connected when entering this routine.
 * 
 */
void otaUpdate(void) {
  char currentSsid[SSID_SZ] = {0};
  char currentPsk[PWD_SZ] = {0};
  char otaSsid[SSID_SZ] = {0};
  char otaPsk[PWD_SZ] = {0};
  boolean needRestart = false;

  log(LOG_DEBUG, S_ATTEMPT_OTA);

  if (!strlen(AUTO_URL)) {
    log(LOG_ERR, S_NO_OTA_URL);
    return;   
  }

  // begin by saving the current WiFi credentials
  
  if (WiFi.SSID().length()) {
    strlcpy(currentSsid, WiFi.SSID().c_str(), SSID_SZ);
    if (WiFi.psk().length()) {
      strlcpy(currentPsk, WiFi.psk().c_str(), PWD_SZ);
    }
  }

  /* The function will try to connect to the following
   * Wi-Fi networks 
   * 
   * 1) OTA_SSID if defined 
   * 2) WiFi.SSID() if defined
   * 3) NET_SSID if defined
   * 
   * to find the firmware file AUTO_URL 
   * 
   */
  if (strlen(OTA_SSID)) {
    strlcpy(otaSsid, OTA_SSID, SSID_SZ);
    strlcpy(otaPsk, OTA_PSK, PWD_SZ);
  } else if (strlen(currentSsid)) {              // WiFi.SSID
    strlcpy(otaSsid, currentSsid, SSID_SZ);
    strlcpy(otaPsk, currentPsk, PWD_SZ);
  } else if (strlen(NET_SSID)) {
    strlcpy(otaSsid, OTA_SSID, SSID_SZ);
    strlcpy(otaPsk, OTA_PSK, PWD_SZ);
  } 
  
  if (!strlen(otaSsid)) {
    log(LOG_ERR, S_NO_OTA_SSID);
    return;
  }
  
  if (connectWiFi(otaSsid, otaPsk)) {
    needRestart = performOtaUpdate(AUTO_URL);
  }  

  /* Restore the initial WiFi network credentials if
   * they have been changed. This is necessary even
   * if the ESP8266 will be restarted which explains
   * why the needRestart variable was introduced.
   */
  if (strcmp(currentSsid, otaSsid)) {
    logf(LOG_DEBUG, S_DISCONNECT_WIFI, WiFi.SSID().c_str());
    WiFi.disconnect();
    if (strlen(currentSsid)) {
      connectWiFi(currentSsid, currentPsk);
    }
  }  
      
  if (needRestart) {
    log(LOG_DEBUG, S_RESTARTING_ESP);
    Serial.flush();
    ESP.restart();
  } 
}

void setup() {
  setLogThreshold(LOG_DEBUG);
  Serial.begin(115200);
  delay(5000);  // enough time for WiFi autoconnect ?

  Serial.println("\n");
  logf(LOG_INFO, S_START_APP, (VERSION >> 16) & 0xFF, (VERSION >> 8) & 0xFF, VERSION & 0xFF);  
  
  stopUartMode();
  
  //log(LOG_DEBUG, S_WAIT_AUTOCONNECT);

  log(LOG_DEBUG, S_LOOK_AT_RESTART_REASON);
  boolean needUpdate = false;
  switch (ESP.getResetInfoPtr()->reason) {
    case REASON_WDT_RST: 
      log(LOG_ERR, S_HARDWARE_WD);
      needUpdate = true;
      break;
      
    case REASON_EXCEPTION_RST: 
      logf(LOG_ERR, S_EXCEPTION_OCCURED, ESP.getResetInfoPtr()->exccause);
      needUpdate = true;
      break;
      
    case REASON_SOFT_WDT_RST:
      log(LOG_ERR, S_SOFTWARE_WD);
      needUpdate = true;
      break;
  }

  if (needUpdate) {
    otaUpdate();
  }
  
  if (WiFi.status() != WL_CONNECTED) {
    connectWiFi(NET_SSID, NET_PSK);
  }
  log(LOG_DEBUG, S_SETUP_DONE);
}

// execute loop() for 3x10 seconds before doing a watch dog timeout

int count = 3;  

void loop() {
  log(LOG_DEBUG, "loop()  ");
  showWiFiStatus();
  // wait 10 seconds
  delay(10000);
  count -= 1;
  if (count <= 0) {
    log(LOG_DEBUG, S_TESTING_WITH_WD);
    while(1) { } // WDT timeout!
  }
}
